/*

HXED, Version 2.00, a Hexadecimal Editor for Binary Files
Copyright 1991 Anthony F. Stuart, All rights reserved.

Date    Name	Change
910209	AFS	Initial release

*/


#include <fcntl.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "sd.h"
#include "def.h"
#include "dcl.h"
#include "mcb.h"
#include "help.h"
#include "proto.h"

MCB	*mcb;

/* Main entry point for HXED */

main (argc, argv)
int	 argc;
char	*argv[];
{
	int	 i;
	MCB	 mcb_buf;

	mcb = &mcb_buf;

	/* Perform hardware and software specific initialization */

	init_hardware ();
	init_software ();

	/* Process command line file name and optional arguments */

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') 
			process_option (argv[i]);
		else
			process_name (argv[i]);
	}

	/* Print usage information and exit if no file name specified */

	if (!mcb->base_path)
		print_usage ();

	/* Initialize screen, display first page and process commands */

	init_screen ();
	display_header ();
	display_page (0l);
	display_byte (0l, cursor_attr);
	display_footer ();
	process_commands ();
}

/* Initialize master control block for file access routines */

init_software ()
{
	mcb->bp_size = DEF_BP_SZ;
	mcb->def_fill = 0;
	mcb->backup = FALSE;
	mcb->base_path = (char *) NULL;
	mcb->temp_path[0] = EOS;
}

/* Process command line options */

process_option (option)
char	*option;
{
	char	 buf[BUF_SZ];

	/* Process each option in the string of options */

	while (*(++option)) {
		switch (*option) {
			case 'b':
				mcb->backup = TRUE;
				return;
			case 'c':
				cursor_attr = get_attr (option+1);
				return;
			case 'd':
				buf[0] = EOS;
				convert_esc_seq (option+1, buf);
				def_disp = buf[0];
				return;
			case 'e':
				buf[0] = EOS;
				convert_esc_seq (option+1, buf);
				mcb->def_fill = buf[0];
				return;
			case 'f':
				footer_attr = get_attr (option+1);
				return;
			case 'h':
				header_attr = get_attr (option+1);
				return;
			case 'm':
				message_attr = get_attr (option+1);
				return;
			case 's':
				silent_mode = TRUE;
				return;
			case 't':
				text_attr = get_attr (option+1);
				return;
			case 'u':
				mcb->bp_size = max (MIN_BP_SZ, atoi (option+1));
				return;
			case 'v':
				set_direct_video (FALSE);
				break;
			default:
				print_usage ();
		}
	}
}

/* Set up a file for editing */

process_name (name)
char	*name;
{
	/* Return if a file is already open */

	if (mcb->base_path)
		return;

	/* Save path name and get last component of path */

	mcb->base_path = name;
	mcb->base_file = get_file_name (name);

	/* Open file and report error if it cannot be opened */

	mcb->base_fd = open (mcb->base_path, O_BINARY|O_RDONLY);
	if (mcb->base_fd == ERR)
		fatal ("Cannot open file");

	/* Create a temporary file for paging */

	create_temp_name (mcb->temp_path);
	mcb->temp_fd = open (mcb->temp_path, O_BINARY|O_CREAT|O_RDWR, 0666);
	if (mcb->temp_fd == ERR)
		fatal ("Cannot open temp file");

	/* Initialize buffer pool and file map */

	init_buffer_pool ();
	init_file_map ();
}

/* Return the last component of a pathname */

char *
get_file_name (name)
char	*name;
{
	char	*last_component;

	last_component = name;

	/* Look for the last DOS or Unix style pathname delimiter */

	while (*name) {
		if (*name == '/' || *name == '\\')
			last_component = name + 1;
		name++;
	}

	/* Truncate to fit on screen if Unix and last component is too long */

	if ((name - last_component) > MAX_NAME_LENGTH)
		last_component = name - MAX_NAME_LENGTH;

	return (last_component);
}

/* Store a value at the current location in the file */

store_data (new_value)
x_char	 new_value;
{
	x_addr	 first_addr;
	x_addr	 last_addr;

	/* If adding a byte beyond end of file then fill in */
	/* the missing bytes between the old end of file and */
	/* the new end of file. */

	if (this_addr > mcb->len) {
		first_addr = max (base_addr, mcb->len);
		last_addr = this_addr + bytes_per_line - 1;
		set_byte_value (this_addr-1, mcb->def_fill);
		update_page (first_addr, last_addr);
	}

	/* If in insert mode insert new byte and then update */
	/* the display. This causes all of the bytes following */
	/* the insert point to wrap to the right. */

	if (insert_mode) {
		first_addr = this_addr;
		insert_byte_value (this_addr, new_value);
		update_page (first_addr, base_addr+bytes_per_page);
		set_position (this_addr, 0);
		return (0);
	}

	/* Otherwise overtype the old value with the new value */

	set_byte_value (this_addr, new_value);
	return (0);
}

/* Manage keystroke interaction with the user */

process_commands ()
{
	int	 key;

	/* Display the signon banner and set an error return */

	display_message ("HXED, Version 2.00, Copyright 1991 Anthony F. Stuart, All rights reserved.");
	setjmp (error_return);

	/* Process commands until terminated by quit command or error */

	for (;;) {

		/* Update the header if necessary */
		
		update_header ();

		/* Get a hex or ascii value, depending on current mode */

		if (hex_mode)
			key = get_hex_input ();
		else
			key = get_ascii_input ();

		/* Clear any messages left over from last command */

		update_message ();

		/* Process command if a command key was entered */

		if (key) 
			process_cmd_key (key);
	}
}

/* Process command entered by user */

process_cmd_key (c)
int	 c;
{
	switch (c) {
		case LEFT:
			move_left ();
			break;
		case UP:
			move_up ();
			break;
		case PGUP:
			page_up ();
			break;
		case RIGHT:
			move_right ();
			break;
		case DOWN:
			move_down ();
			break;
		case PGDOWN:
			page_down ();
			break;
		case HOME:
			set_position (0l, 0);
			break;
		case END:
			set_position (mcb->len, bytes_per_page/2);
			break;
		case GOTO:
			goto_addr ();
			break;
		case DELETE:
			delete_char ();
			break;
		case INSERT:
			insert_mode ^= ~0;
			break;
		case SEARCH:
			get_pattern_and_search ();
			break;
		case REPLACE:
			search_and_replace ();
			break;
		case SUSPEND:
			suspend_to_os ();
			break;
		case WRITE:
			get_name_and_write ();
			break;
		case HEX:
			hex_mode ^= ~0;
			break;
		case QUIT:
			quit ();
			break;
		default:
			help ();
			break;
	}
}

/* Move the cursor one position to the left */

move_left ()
{
	x_addr	 old_base_addr;

	/* Return if cursor is already at home */

	if (this_addr == 0)
		return;

	/* Turn off the cursor attribute and decrement the current addr */

	display_byte (this_addr, text_attr);
	this_addr--;

	/* If the current address is on the previous page then scroll up */

	if (this_addr < base_addr) {
		old_base_addr = base_addr;
		base_addr -= bytes_per_line;
		scroll_screen_down ();
		update_page (base_addr, old_base_addr);
	}

	/* Turn on the cursor attribute */

	display_byte (this_addr, cursor_attr);
}

/* Move the cursor up one row, equivalent to moving it left 16 bytes */

move_up ()
{
	x_addr	 old_base_addr;

	/* Return if cursor is already at home */

	if (this_addr < bytes_per_line)
		return;

	/* Turn off the cursor attribute and decrement the current addr */

	display_byte (this_addr, text_attr);
	this_addr -= bytes_per_line;

	/* If the current address is on the previous page then scroll up */

	if (this_addr < base_addr) {
		old_base_addr = base_addr;
		base_addr -= bytes_per_line;
		scroll_screen_down ();
		update_page (base_addr, old_base_addr);
	}

	/* Turn on the cursor attribute */

	display_byte (this_addr, cursor_attr);
}

/* Move up one page */

page_up ()
{
	/* Turn off the cursor attribute */

	display_byte (this_addr, text_attr);

	/* Set the current addr and the screen base addr back one page */

	this_addr = sub_addr (this_addr, bytes_per_page);
	base_addr = sub_addr (base_addr, bytes_per_page);

	/* Display the new page and turn the cursor back on */

	display_page (base_addr);
	display_byte (this_addr, cursor_attr);
}

/* Move the cursor one position to the right */

move_right ()
{
	x_addr	 end_addr;

	/* Turn off the cursor attribute and increment the current addr */

	display_byte (this_addr, text_attr);
	end_addr = base_addr + bytes_per_page;
	this_addr++;

	/* If the current address is on the next page then scroll down */

	if (this_addr >= end_addr) {
		base_addr += bytes_per_line;
		scroll_screen_up ();
		update_page (end_addr, end_addr+bytes_per_line);
	}

	/* Turn on the cursor attribute */

	display_byte (this_addr, cursor_attr);
}

/* Move the cursor down one row, equivalent to moving it right 16 bytes */

move_down ()
{
	x_addr	 end_addr;

	/* Turn off the cursor attribute and increment the current addr */

	display_byte (this_addr, text_attr);
	end_addr = base_addr + bytes_per_page;
	this_addr += bytes_per_line;

	/* If the current address is on the next page then scroll down */

	if (this_addr >= end_addr) {
		base_addr += bytes_per_line;
		scroll_screen_up ();
		update_page (end_addr, end_addr+bytes_per_line);
	}

	/* Turn on the cursor attribute */

	display_byte (this_addr, cursor_attr);
}

/* Move down one page */

page_down ()
{
	/* Turn off the cursor attribute */

	display_byte (this_addr, text_attr);

	/* Set the current addr and the screen base addr forward one page */

	this_addr += bytes_per_page;
	base_addr += bytes_per_page;

	/* Display the new page and turn the cursor back on */

	display_page (base_addr);
	display_byte (this_addr, cursor_attr);
}

/* Delete a character at the current cursor position */

delete_char ()
{
	delete_byte_addr (this_addr);
	update_page (this_addr, base_addr+bytes_per_page);
	display_byte (this_addr, cursor_attr);
}

/* Get and address from the user and goto that address */

goto_addr ()
{
	char	 buf[BUF_SZ];

	buf[0] = EOS;
	get_text_input ("Goto Addr: ", buf, BUF_SZ);
	set_position (strtol (buf, (char **) NULL, 0), bytes_per_page / 2);
}

/* Find the best way to move from this_addr to next_addr */

set_position (addr, context)
x_addr	 addr;
int	 context;
{
	x_addr	 this_end_addr;
	x_addr	 prev_end_addr;
	x_addr	 next_end_addr;

	/* Get the end of this page, the next page, and the previous page */

	this_end_addr = base_addr + bytes_per_page;
	next_end_addr = this_end_addr + bytes_per_page;
	prev_end_addr = sub_addr (base_addr, bytes_per_page);

	/* Turn off cursor attributes if old address is on this page */

	if ((this_addr >= base_addr) && (this_addr < this_end_addr))
		display_byte (this_addr, text_attr);

	/* If addr is on next page then scroll down to it */

	if ((addr >= this_end_addr) && (addr < next_end_addr)) {
		while (addr >= this_end_addr) {
			base_addr += bytes_per_line;
			scroll_screen_up ();
			update_page (this_end_addr, this_end_addr+bytes_per_line);
			this_end_addr += bytes_per_line;
		}
	}

	/* If addr is on previous page then scroll up to it */

	else if ((addr < base_addr) && (addr >= prev_end_addr)) {
		while (addr < base_addr) {
			base_addr -= bytes_per_line;
			scroll_screen_down ();
			update_page (base_addr, base_addr+bytes_per_line);
		}
	}

	/* If addr is far away then redraw screen */

	else if ((addr < base_addr) || (addr > this_end_addr)) {
		base_addr = get_row_addr (sub_addr (addr, context));
		display_page (base_addr);
	}

	/* Make addr the new current addr and highlight it */

	this_addr = addr;
	display_byte (this_addr, cursor_attr);

}

/* Suspend HXED and invoke an OS shell */

suspend_to_os ()
{
	/* Clear the editor display and tell user how to get back */

	clear_screen (DEFAULT_ATTR);
	display_string (minrow+1, mincol, "Type EXIT to return to HXED...", DEFAULT_ATTR);
	position_cursor (minrow+2, mincol);
	restore_hardware ();

	/* Invoke the OS shell, it will remain active until terminated */

	system (SHELL);

	/* Restore the editor display and return to editing */

	init_hardware ();
	init_screen ();
	display_header ();
	display_page (base_addr);
	display_byte (this_addr, cursor_attr);
	display_footer ();
}

/* Process the write command */

get_name_and_write ()
{
	int	 rc;
	char	 new_name[BUF_SZ];
 
	struct stat	 sb;

	/* Give user a chance to back out */

	if (get_response ("Write file? (yes/no) ", "yn") == 'n')
		return;

	/* Get file name and write file to it */

	new_name[0] = EOS;
	get_text_input ("File: ", new_name, BUF_SZ);
	if (new_name[0] == EOS)
		return;

	write_file (new_name);

	/* See if base file has been updated */

	rc = stat (mcb->base_path, &sb);
	if (rc == ERR)
		fatal ("Cannot stat base file");

	if (sb.st_mtime == mcb->base_mtime)
		return;

	/* If so, re-initialize buffer pool and file map */

	re_init_buffer_pool ();
	re_init_file_map ();
	display_footer ();
}

/* Write the current file structure to a file */

write_file (path_name)
char	*path_name;
{
	int	 rc;
	char	 temp_name[BUF_SZ];

	/* Create a temporary file and write blocks to it */

	validate_path_name (path_name);
	create_temp_name (temp_name);
	write_file_map (temp_name);
	dispose_of_file (path_name);

	/* Rename temporary file to new name specified by user */

	rc = rename (temp_name, path_name);
	if (rc == ERR)
		fatal ("Cannot rename file");

	mcb->change = FALSE;
}

/* Get rid of a pre-existing output file */

dispose_of_file (path_name)
char	*path_name;
{
	char	*tp;
	char	*sp;
	char	*dot;
	char	 backup_name[BUF_SZ];

	/* Just unlink it if user doesn't want a backup copy */

	if (!mcb->backup) {
		unlink (path_name);
		return;
	}

	/* Otherwise find the extension of the last path name component */

	tp = backup_name;
	sp = path_name;
	dot = (char *) NULL;

	while (*sp) {
		if (*sp == '/' || *sp == '\\')
			dot = (char *) NULL;
		if (*sp == '.')
			dot = tp;
		*tp++ = *sp++;
	}

	/* Start at dot if filename has an extension */

	if (dot)
		tp = dot;

	/* And append a .bak suffix to it */

	sp = ".bak";
	while (*sp) {
		*tp++ = *sp++;
	}

	*tp = EOS;

	/* Unlink any previous file by that name and rename ours */

	unlink (backup_name);
	rename (path_name, backup_name);
}

/* Process the quit command */

quit ()
{
	/* Give user a chance to back out */

	if (get_response ("Quit? (yes/no) ", "yn") == 'n')
		return;

	/* Write file if it has been updated and user wants to save changes */

	if (mcb->change) {
		if (get_response ("Save changes? (yes/no) ", "yn") == 'y')
			write_file (mcb->base_path);
	}

	/* Clean up and exit */
		
	unlink (mcb->temp_path);					
	clear_screen (DEFAULT_ATTR);
	restore_hardware ();
	exit (0);
}

/* Process the help command */

help ()
{
	int	 i;

	/* Display the help text over the current editor display */

	for (i = 0; i < HELP_SZ; i++) {
		display_string (help_row+i, help_col, help_text[i], cursor_attr);
	}

	display_message ("HXED Help - Press Any Key to Continue");

	/* Wait for the user to enter a key, then redraw the display */

	get_keystroke (CMD_OR_DATA);
	display_page (base_addr);
	display_byte (this_addr, cursor_attr);
	update_message ();
}		


